# coding=utf-8

from config import *
from wrapper_mysql import MySQL
from wrapper_sqlite3 import SQLite3
from exception import MonSQLException

class DB_TYPES:
	MYSQL = 'MySQL'
	SQLITE3 = 'SQLite3'

def MonSQL(host=None, port=None, username=None, password=None, dbname=None, dbpath=None, dbtype=None):
	if dbtype is None:
		raise MonSQLException('Database type must be specified')

	if dbtype == DB_TYPES.MYSQL:
		return MySQL(host, port, username, password, dbname)
	elif dbtype == DB_TYPES.SQLITE3:
		return SQLite3(dbpath)
	else:
		raise MonSQLException('Database type %s not supported' %dbtype)


# coding=utf-8

from query import Query, QueryCondition, value_to_sql_str
from datetime import datetime, date
import types
from exception import MonSQLException
import uuid
from config import ASCENDING, DESCENDING

def from_none_to_null(v):
    if v is not None: return v
    return u'null'

def build_query(query):
    return QueryCondition(query).to_sql()

def build_select_query(table_name, values, query, sort=None, skip=0, limit=None, distinct=False):

    value_str = ""
    for index, field in enumerate(values):
        if field in ('index',):
            field = "'%s'" %(field)
            
        if index == 0:
            value_str = field
        else:
            value_str += u"," + field

    # Where clause
    query_str = build_query(query)
    if query_str:
        query_str = u"WHERE " + query_str
    else:
        query_str = ""

    if distinct:
        value_str = 'DISTINCT %s' %value_str

    sql = u"""SELECT %s FROM %s %s""" %(value_str, table_name, query_str)

    if sort:
        sort_strings = []

        for item in sort:
            assert len(item) == 2
            assert item[1] in (ASCENDING, DESCENDING)
            sort_str = '%s ' %(item[0])
            if item[1] == ASCENDING:
                sort_str += 'ASC'
            elif item[1] == DESCENDING:
                sort_str += 'DESC'
            sort_strings.append(sort_str)

        sql = '''%s ORDER BY %s''' %(sql, ','.join(sort_strings))

    if limit is not None:
        sql = '%s LIMIT %d,%d' %(sql, skip, limit)

    return sql


def build_select(query_obj):
    """
    Given a Query obj, return the corresponding sql
    """
    return build_select_query(query_obj.source, query_obj.fields, query_obj.filter, skip=query_obj.skip, \
                              limit=query_obj.limit, sort=query_obj.sort, distinct=query_obj.distinct)


def build_insert(table_name, attributes):
    """
    Given the table_name and the data, return the sql to insert the data
    """
    sql = "INSERT INTO %s" %(table_name)
    column_str = u""
    value_str = u""
    for index, (key, value) in enumerate(attributes.items()):
        if index > 0:
            column_str += u","
            value_str += u","
        column_str += key
        value_str += value_to_sql_str(value)
    sql = sql + u"(%s) VALUES(%s)" %(column_str, value_str)
    return sql


def build_delete(table_name, condition):
    query_str = build_query(condition)
    if not query_str:
        query_str = u""
    else:
        query_str = u"WHERE " + query_str
    sql = u"DELETE FROM %s %s" %(table_name, query_str)
    return sql


def build_update(table_name, condition, attributes):
    sql = u"UPDATE %s SET " %(table_name)    
    set_str = u""
    for index, (key, value) in enumerate(attributes.items()):
        if index > 0:
            set_str += u","
        set_str += key + u"=" + value_to_sql_str(value)

    query_str = build_query(condition)
    if query_str:
        query_str = u" WHERE " + query_str
    else:
        query_str = u""
    sql = sql + set_str + query_str
    return sql



# coding=utf-8
import abc
from logging import Logger
from config import TRANSACTION_MODE
from query import Query
from queryset import QuerySet
from sql import build_query, build_select, build_update, build_delete, build_insert
  

class Table:
    """
    A collection of related data entries in columns and rows.
    This class should not be directly constructed. Instead use
    Database.get('table_name') to have a table returned.
    """

    db = None
    cursor = None
    columns = None
    name = None

    def __init__(self, db, name, mode=None):
        self.db = db
        self.cursor = db.cursor()
        self.columns = None
        self.name = name

        if mode:
            self.transaction_mode = mode
        else:
            self.transaction_mode = TRANSACTION_MODE.DEFAULT

        self.logger = Logger("std")


    def columns():
        doc = "The columns property."
        def fget(self):
            return self._columns
        def fset(self, value):
            self._columns = value
        def fdel(self):
            del self._columns
        return locals()

    columns = property(**columns())


    @abc.abstractmethod
    def fetch_columns(self):
        pass


    def __ensure_columns(self):
        if self.columns is not None:
            return True
        self.fetch_columns()
        return True


    def commit(self):
        """
        Ends current transaction, making permanent any changes made.
        """
        self.db.commit()
        return self
    

    def count(self, query=None, distinct=False, distinct_fields=None):
        """
        Returns the number of rows satisying a criteria, if provided.
        
        :Parameters: 

        - query: specify the WHERE clause
        - distinct : boolean, whether to use DISTINCT()
        - distinct_fields : list or tuple or strings. Each string is
          a column name used inside COUNT(). If none, '*' will be
          used.

        :Return: int, the number of rows
        """
        if distinct_fields is None:
            self.__ensure_columns()
            if distinct:
                field = ','.join(self.columns)
            else:
                field = '*'
        else:
            field = ','.join(distinct_fields)

        if distinct:
            count_str = 'DISTINCT(%s)' %(field)
        else:
            count_str = field

        count_str = 'COUNT(%s)' %(count_str)

        sql = 'SELECT %s FROM %s' %(count_str, self.name)

        if query is not None:
            query_str = build_query(query)
            if query_str:
                sql = sql + ' WHERE ' + query_str

        self.cursor.execute(sql)
        count = self.cursor.fetchone()[0]

        return count


    def find(self, filter={}, fields=None, skip=0, limit=None, sort=None):
        """
        Searches the table using the filters provided.
        
        :Examples:

        >>> users = user_table.find({'id': {'$in': [10, 20]}, 'age': {'$gt': 20}}) # Complex query
        >>> user_count = len(users)
        >>> for user in users:
        >>>     # Do something...
        >>>     print user.id
        >>> 
        >>> users = user_table.find({}, sort=[('age', monsql.ASCENDING)]) # sort by age

        Also support complex operators:

        >>> {a: 1}                                  # a == 1
        >>> {a: {$gt: 1}}                           # a > 1
        >>> {a: {$gte: 1}}                          # a >= 1
        >>> {a: {$lt: 1}}                           # a < 1
        >>> {a: {$lte: 1}}                          # a <= 1
        >>> {a: {$eq: 1}}                           # a == 1
        >>> {a: {$in: [1, 2]}}                      # a == 1 or a == 2
        >>> {a: {$contains: '123'}}                 # a like %123%
        >>> {$not: condition}                       # !(condition)
        >>> {$and: [condition1, condition2, ...]}   # condition1 and condition2
        >>> {$or: [condition1, condition2, ...]}    # condition1 or condition2

        :Parameters: 

        - query(dict): specify the WHERE clause. One example is {"name": "...", "id": ...}    
        - fields: specify what fields are needed
        - skip, limit: both integers, skip without defining limit is meaningless
        - sort: A list, each element is a two-item tuple, with the first item be the column name
          and the second item be either monsql.ASCENDING or monsql.DESCENDING

        :Return: a QuerySet object
        """
        if not fields:
            self.__ensure_columns()
            fields = self.columns

        query_obj = Query(source=self.name, filter=filter, fields=fields, skip=skip, limit=limit, sort=sort)
        return QuerySet(cursor=self.cursor, query=query_obj)
    

    def find_one(self, filter=None, fields=None, skip=0, sort=None):
        """
        Similar to find. This method will only retrieve one row.
        If no row matches, returns None
        """
        result = self.find(filter=filter, fields=fields, skip=skip, limit=1, sort=sort)
        if len(result) > 0:
            return result[0]
        else:
            return None


    def insert(self, data_or_list_of_data):
        """
        Insert data into the table.
        
        :Examples:
        
        >>> user_table.insert({'username': 'Jude'}) # Insert one row
        >>> user_table.insert([{'username': 'Andy'}, {'username': 'Julia'}, ...]) # Insert multiple rows

        :Parameters:

        - data_or_list_of_data: Either a dict or a list of dict

        :Return: id or list of ids of inserted row
        """
        result = None

        def insert_data(data):
            sql = build_insert(self.name, data)
            row_count = self.cursor.execute(sql)

            if row_count:
                return self.cursor.lastrowid
            else:
                return None

        value_to_be_inserted = []

        if isinstance(data_or_list_of_data, list):
            result = []
            for data in data_or_list_of_data:
                r = insert_data(data)
                result.append(r)
        else:
            result = insert_data(data_or_list_of_data)
        
        return result
    

    def update(self, query, attributes, upsert=False):
        """
        Updates data in the table.
        
        :Parameters: 

        - query(dict), specify the WHERE clause
        - attributes(dict), specify the SET clause
        - upsert: boolean. If True, then when there's no row matches the query, insert the values

        :Return: Number of rows updated or inserted
        """
        if upsert:
            found_result = self.find_one(query)
            if not found_result:
                id = self.insert(attributes)
                if id > 0:
                    return 1
                else:
                    return 0

        sql = build_update(self.name, query, attributes)
        return self.cursor.execute(sql)
    
    
    def remove(self, filter=None):
        """
        Removes rows from the table.
        
        :Parameters: 

        - query(dict), specify the WHERE clause

        :Return: Number of rows deleted
        """
        sql = build_delete(table_name=self.name, condition=filter)
        return self.cursor.execute(sql)




'''
SQLite3
-------
A lightweight disk-based database. Usually useful for internal
data storage and prototyping for applications.
'''


import sqlite3
from db import Database
from table import Table

class SQLite3Table(Table):
    def fetch_columns(self):
        sql = u"PRAGMA table_info(%s)" %(self.name)
        self.cursor.execute(sql)
        columns = []

        for column in self.cursor.fetchall():
            column = column[1]
            columns.append(column)
        self.columns = columns


class SQLite3(Database):

    def __init__(self, file_path=None):
        if file_path is None: file_path = ":memory:"
        db = sqlite3.connect(file_path)
        Database.__init__(self, db)

    def list_tables(self):
        self.cursor.execute('''SELECT name FROM sqlite_master WHERE type = 'table' ''')
        all_tablenames = [row[0].lower() for row in self.cursor.fetchall()]
        return all_tablenames

    def get_table_obj(self, name):
        table = SQLite3Table(db=self.db, name=name, mode=self.mode)
        return table

    def truncate_table(self, tablename):
        """
        SQLite3 doesn't support direct truncate, so we just use delete here
        """
        self.get(tablename).remove()
        self.db.commit()

        


'''
MySQL
--------
The World's "most" popular open source database
'''


import MySQLdb
from db import Database
from table import Table
from config import TRANSACTION_MODE

class MySQLTable(Table):
    
    def fetch_columns(self):
        sql = u"SHOW COLUMNS FROM %s" %(self.name)
        self.cursor.execute(sql)
        columns = []
        for column in self.cursor.fetchall():
            column = column[0]
            columns.append(column)
        self.columns = columns


class MySQL(Database):

    def __init__(self, host='127.0.0.1', port=3306, username='', password='',
                 dbname='test', mode=TRANSACTION_MODE.DEFAULT):
        db = MySQLdb.Connect(host=host, port=port, user=username,
                                    passwd=password, db=dbname)
        Database.__init__(self, db)

    def list_tables(self):
        self.cursor.execute('show tables')
        all_tablenames = [row[0].lower() for row in self.cursor.fetchall()]
        return all_tablenames

    def get_table_obj(self, name):
        table = MySQLTable(db=self.db, name=name, mode=self.mode)
        return table

    def truncate_table(self, tablename):
        """
        Use 'TRUNCATE TABLE' to truncate the given table
        """
        self.cursor.execute('TRUNCATE TABLE %s' %tablename)
        self.db.commit()

        


import unittest
import sys
from datetime import *
import uuid
import time
from base import *
from monsql import MonSQL, DESCENDING, DB_TYPES
from monsql.exception import MonSQLException


class MonSQLBasicTest(BaseTestCase):
    
    def setUp(self):
        super(MonSQLBasicTest, self).setUp()

        test_table_name_a = 'test_a_' + random_name()

        self.monsql.create_table(test_table_name_a, ['name VARCHAR(50)', 'number INT', 'datetime DATETIME', 'date DATE', 'double_number DOUBLE'])
        self.table_a = self.monsql.get(test_table_name_a)

    def tearDown(self):
        self.monsql.drop_table(self.table_a.name)
        super(MonSQLBasicTest, self).tearDown()

    def test_delete_all(self):
        self._insert_some_row_to_table_one(10)
        assert self.table_a.count() == 10       

        self.table_a.remove()
        self.monsql.commit()

        assert self.table_a.count() == 0

    def _insert_some_row_to_table_one(self, count):
        for i in range(count):
            self.table_a.insert({"name": "jude" + str(i), "number": i, "datetime": datetime.now(), "date": date.today(), "double_number": 4.33333})
        self.monsql.commit()

    def test_complex_queries(self):
        self._insert_some_row_to_table_one(10)

        self.assertEquals(len(self.table_a.find({})), 10)

        filters_and_expected_row_nums = [
            ({'number': {'$gte': 1}}, 9),
            ({'number': {'$gt': 1}}, 8),
            ({'number': {'$lt': 1}}, 1),
            ({'number': {'$lte': 1}}, 2),
            ({'number': {'$eq': 1}}, 1),
            ({'number': {'$in': [1, 11]}}, 1),
            ({'name': {'$contains': 'de0'}}, 1),
        ]

        for filter, row_num in filters_and_expected_row_nums:
            self.assertEqual(len(self.table_a.find(filter)), row_num)
            self.assertEqual(len(self.table_a.find({'$not': filter})), 10 - row_num)

        composite_filters_and_expected_row_nums = [
            ({'$and': [{'name': 'jude0'}, {'number': 0}]}, 1),
            ({'$and': [{'name': 'jude0'}, {'number': 1}]}, 0),
            ({'$or': [{'name': 'jude0'}, {'number': 0}]}, 1),
            ({'$or': [{'name': 'jude0'}, {'number': 1}]}, 2),
        ]

        for filter, row_num in composite_filters_and_expected_row_nums:
            self.assertEqual(len(self.table_a.find(filter)), row_num)


    def test_insert(self):
        # Test findall, findone
        self._insert_some_row_to_table_one(10)

        result = self.table_a.find()
        self.assertTrue(len(result) == 10)
        result_one = self.table_a.find_one()
        self.assertTrue(result_one is not None)

        # Test conditioned query
        for i in range(10):
            result_conditioned = self.table_a.find({"name": "jude" + str(i)})
            self.assertTrue(len(result_conditioned) == 1)

            result_conditioned = self.table_a.find({"number": i})
            self.assertTrue(len(result_conditioned) == 1)

    def test_sorting(self):
        self._insert_some_row_to_table_one(10)

        data = self.table_a.find({}, sort=[('number', DESCENDING)])
        first_row = self.table_a.find_one(sort=[('number', DESCENDING)])
        for i in range(len(data) - 1):
            self.assertTrue(data[i].number > data[i + 1].number)

        assert first_row.number == max([r.number for r in data])

    def test_delete(self):
        self._insert_some_row_to_table_one(10)

        for i in range(10):
            self.table_a.remove({"number": i})
            self.table_a.commit()
            all_result = self.table_a.find()
            self.assertTrue(len(all_result) == 10 - i - 1)

    def test_update(self):
        self._insert_some_row_to_table_one(10)

        for i in range(10):
            self.table_a.update(query={"number": i}, attributes={"number": i - 100})
            self.assertTrue(len(self.table_a.find({"number": i})) == 0)
            self.assertTrue(len(self.table_a.find({"number": i - 100})) == 1)

    def test_limit(self):
        self._insert_some_row_to_table_one(10)

        full_rows = self.table_a.find().values()

        rows_with_limit = self.table_a.find(limit=5).values()
        rows_with_limit_and_skip = self.table_a.find(limit=5, skip=5).values()

        self.assertEqual(full_rows[: 5], rows_with_limit)
        self.assertEqual(full_rows[5: ], rows_with_limit_and_skip)

    def test_query_set(self):
        self._insert_some_row_to_table_one(10)
        self._insert_some_row_to_table_one(10)

        query_set = self.table_a.find()
        self.assertEqual(query_set.count, 20)
        self.assertEqual(query_set.distinct().count, 10)
        self.assertEqual(query_set.filter({'number': 1}).count, 2)
        self.assertEqual(query_set.filter({'number': 1}).filter({'name': 'jude1'}).count, 2)

        self.assertEqual(query_set.filter({'number': 1}).filter({'name': 'jude1'}).distinct().count, 1)
        self.assertEqual(query_set.filter({'number': 1}).filter({'name': 'jude0'}).count, 0)

        self.assertEqual(self.table_a.find(fields=['date']).distinct().count, 1)

    def test_raw(self):
        self._insert_some_row_to_table_one(10)
        rows = self.monsql.raw('select * from %s' %self.table_a.name)
        for idx, row in enumerate(rows):
            self.assertEqual(row.number, idx)
        
        self.monsql.raw('update %s set number=100000' %self.table_a.name)
        self.monsql.commit()

        self.assertEqual(self.table_a.find({'number': 1}).count, 0)
        self.assertEqual(self.table_a.find({'number': 100000}).count, 10)
        self.monsql.commit()






# coding=utf-8

import copy, types
from datetime import datetime, date
from exception import MonSQLException


class Query:
    def __init__(self, source=None, filter=None, fields=None, skip=0, limit=None, sort=None, distinct=False, alias=None):
        self.source = source
        self.filter = filter
        self.fields = fields
        self.sort = sort
        self.limit = limit
        self.skip  = skip
        self.distinct = distinct
        self.alias = alias

    def clone(self):
        return Query(source=copy.deepcopy(self.source), 
                 filter=copy.deepcopy(self.filter), 
                 fields=copy.deepcopy(self.fields), 
                 skip=self.skip, 
                 limit=self.limit,
                 sort=copy.deepcopy(self.sort),
                 distinct=self.distinct,
                 alias=copy.deepcopy(self.alias))   

    def add_filter(self, obj):
        self.filter = {'$and': [obj, self.filter]}


def value_to_sql_str(v):
    if v is None:
        return 'null'

    if type(v) in (types.IntType, types.FloatType, types.LongType):
        return str(v)

    if type(v) in (types.StringType, types.UnicodeType):
        return "'%s'" %(v.replace(u"'", u"\\'"))

    if isinstance(v, datetime):
        return "'%s'" %(v.strftime("%Y-%m-%d %H:%M:%S"))

    if isinstance(v, date):
        return "'%s'" %(v.strftime("%Y-%m-%d"))

    return str(v)


class QueryCondition:

    MYSQL_RESERVE_WORDS                 = (u'index', )
    COMPLEX_QUERY_INDICATOR             = (u'$not', u'$and', u'$or')


    def __init__(self, condition):
        self.condition = condition

    def to_sql(self):
        """
        This function build a sql condition string (those used in the 'WHERE' clause) based on given condition
        Supported match pattern:

        {a: 1}                              -> a == 1
        {a: {$gt: 1}}                       -> a > 1
        {a: {$gte: 1}}                      -> a >= 1
        {a: {$lt: 1}}                       -> a < 1
        {a: {$lte: 1}}                      -> a <= 1
        {a: {$eq: 1}}                       -> a == 1
        {a: {$in: [1, 2]}}                  -> a == 1
        {a: {$contains: '123'}}             -> a like %123%

        And complex combination
        {$not: condition}                   -> !(condition)
        {$and: [condition1, condition2]}    -> condition1 and condition2
        {$or: [condition1, condition2]}     -> condition1 or condition2

        """
        condition = self.condition
        if condition is not None and len(condition.items()) > 0:
            keys = condition.keys()
            if len(keys) > 1:
                split_conditions = []
                for key in condition.keys():
                    split_conditions.append({key: condition[key]})

                return QueryCondition({'$and': split_conditions}).to_sql()
            else:
                query_field, query_value = condition.items()[0]

                if query_field in QueryCondition.COMPLEX_QUERY_INDICATOR:

                    if u'$not' == query_field:
                        not_condition = QueryCondition(query_value).to_sql()
                        if not_condition is not None:
                            return 'NOT (%s)' %(not_condition)
                        else:
                            return None

                    if query_field in (u'$or', u'$and', ):
                        conditions = query_value
                        if not isinstance(conditions, list) or len(conditions) < 2:
                            raise MonSQLException('QUERY VALUE FOR KEY %s MUST BE LIST WITH LENGTH BEING AT LEAST 2' %(query_field))

                        conditions = map(lambda c: QueryCondition(c).to_sql(), conditions)
                        conditions = filter(lambda c: c is not None, conditions)

                        if len(conditions) > 0:
                            if query_field == u'$or':
                                return ' OR '.join(conditions)
                            elif query_field == u'$and':
                                return ' AND '.join(conditions)
                        else:
                            return None
                    else:
                        raise Exception('Unsupport query_field')
                else:
                    if query_field in QueryCondition.MYSQL_RESERVE_WORDS:
                        query_field = "'%s'" %(query_field)

                    if not type(query_value) is types.DictType:
                        query_value = {'$eq': query_value}

                    if len(query_value.keys()) > 1:
                        # Deal with situation like a: {'$gt': 1, '$lt': 10}
                        # Split into {$and: [a: {'$gt': 1}, a: {'$lt': 10}]}
                        split_conditions = []
                        for key in query_value.keys():
                            split_conditions.append(QueryCondition({query_field: {key: query_value[key]}}))

                        return QueryCondition({'$and': split_conditions}).to_sql()
                    else:
                        # The simple case of {a: {$eq: 1}}
                        match_key = query_value.keys()[0]
                        match_value = query_value[match_key]

                        query_str = None
                        if u"$contains" == match_key:
                            query_str = u"LIKE " + value_to_sql_str('%' + match_value + '%')

                        elif match_key in ('$eq', '$gte', '$gt', '$lt', '$lte'):
                            map_dic = {'$eq': '=', '$gte': '>=', '$gt': '>', '$lt': '<', '$lte': '<='}
                            query_str = map_dic[match_key] + value_to_sql_str(match_value)

                        elif u'$in' == match_key:
                            if len(match_value) == 0:
                                query_str = u"IN (null) "
                            else:
                                query_str = u"IN (" + u','.join([str(_v_) for _v_ in match_value]) + u") "
                        else:
                            raise MonSQLException(u"Unsupport complex query: %s" %(match_key))

                        return query_field + ' ' + query_str
        else:
            return None

        # For testing
        assert False



# coding=utf-8

from query import Query
from sql import build_select
from exception import MonSQLException

class DataRow:

    def __init__(self, keyvalue_map):
        self.__data = keyvalue_map

    def __getattr__(self, attrname):

        if self.data.has_key(attrname):
            return self.data[attrname]
        else:
            raise AttributeError('%s does not exist' %attrname)

    @property
    def data(self):
        return self.__data


class QuerySet:
    """
    Lazy load data
    """

    def __init__(self, cursor, query):
        self.cursor = cursor
        self.query = query
        self._data = None
        self._need_to_refetch_data = False

    """
    Python magic functions
    """
    def __len__(self):
        return self.count

    def __iter__(self):
        if not self._data or self._need_to_refetch_data:
            self._fetch_data()
        return iter(self._data)

    def __getitem__(self, k):
        if not self._data or self._need_to_refetch_data:
            self._fetch_data()
        return self._data[k]


    @property
    def count(self):
        if not self._data or self._need_to_refetch_data:
            self._fetch_data()
        return len(self._data)

    def _fetch_data(self):

        sql = build_select(self.query)
        # print sql
        self.cursor.execute(sql)

        data_list = self.cursor.fetchall()
        values = self.query.fields
        result_list = []
        for data in data_list:
            assert(len(data) == len(values))
            result = {}
            for i in range(len(data)):
                result[values[i]] = data[i]
            result_list.append(DataRow(result))

        self._data = result_list
        self._need_to_refetch_data = False

    def filter(self, filter):
        """
        Add new filter to the query set. Note that since QuerySet is lazy, it would
        just combine this filter with the original filter with an 'AND'
        :Parameters: 
        - filter: a dictionary
        :Return: a new QuerySet object
        """
        new_query_set = self.clone()
        new_query_set.query.add_filter(filter)
        
        return new_query_set


    def limit(self, n, skip=None):
        """
        Limit the result set. However when the query set already has limit field before,
        this would raise an exception
        :Parameters:
        - n : The maximum number of rows returned
        - skip: how many rows to skip
        :Return: a new QuerySet object so we can chain operations
        """
        if self.query.limit is not None:
            raise MonSQLException('LIMIT already defined')

        new_query_set = self.clone()
        new_query_set.query.limit = n
        new_query_set.query.skip = skip

        return new_query_set


    def distinct(self):
        """
        Only return distinct row. 
        Return a new query set with distinct mark
        """
        new_query_set = self.clone()
        new_query_set.query.distinct = True
        return new_query_set


    def clone(self):
        return QuerySet(cursor=self.cursor, query=self.query.clone())


    def exists(self):
        return len(self) > 0


    def sort(self, sort):
        self._need_to_refetch_data = True
        raise Exception('NOT IMPLEMENTED')


    def values(self):
        return [v.data for v in self]




import unittest
import sys
from datetime import *
import time
from base import *
from monsql import MonSQL, DESCENDING, DB_TYPES
from monsql.exception import MonSQLException

class MonSQLMetaOperationTest(BaseTestCase):
    
    def test_table_operations(self):
        tablename = 'test_table' + random_name()
        self.monsql.create_table(tablename, [('id INT NOT NULL')])
        self.assertTrue(self.monsql.is_table_existed(tablename))
        table_obj = self.monsql.get(tablename)

        for id in range(10):
            table_obj.insert({'id': id})
        self.monsql.commit()

        self.assertTrue(table_obj.count() == 10)
        for id in range(10):
            table_obj.insert({'id': id})

        self.assertEqual(table_obj.count(), 20)
        self.assertEqual(table_obj.count(distinct=True), 10)
        self.assertEqual(table_obj.count(query={'id': 2}), 2)
        self.assertEqual(table_obj.count(distinct=True, query={'id': 2}), 1)

        self.monsql.truncate_table(tablename)
        self.assertTrue(table_obj.count() == 0)

        is_exception_raised = False
        
        try:
            self.monsql.create_table(tablename, [('id INT')], force_recreate=False)
        except MonSQLException as e:
            is_exception_raised = True
        self.assertTrue(is_exception_raised)

        self.monsql.create_table(tablename, [('id INT')], primary_key=['id'], force_recreate=True)
        self.assertTrue(self.monsql.is_table_existed(tablename))

        self.monsql.drop_table(tablename)
        self.assertFalse(self.monsql.is_table_existed(tablename))


import unittest
import uuid
from monsql import MonSQL, DB_TYPES
import yaml
import os

def random_name():
    uids = str(uuid.uuid1()).split('-')
    return uids[1] + uids[3];

def load_test_settings():
    config_path = os.path.join(os.path.join(os.path.dirname(__file__), 'config.yml'))
    if os.path.exists(config_path):
        return yaml.load(open(config_path).read())
    else:
        return os.environ

class BaseTestCase(unittest.TestCase):
    # STATIC
    DB_TYPE = None

    # INSTANCE VARIABLES
    monsql = None

    def setUp(self):
        dbtype = os.environ['DB_TYPE']
        config_data = load_test_settings()

        if dbtype == DB_TYPES.MYSQL:
            self.monsql = MonSQL(host=config_data['MYSQL_HOST'], 
                                 port=config_data['MYSQL_PORT'], 
                                 username=config_data['MYSQL_USER'],
                                 password=config_data['MYSQL_PASSWORD'], 
                                 dbname=config_data['MYSQL_DB'], 
                                 dbtype=dbtype)

        elif dbtype == DB_TYPES.SQLITE3:
            self.monsql = MonSQL(dbpath=config_data['SQLITE3_PATH'],
                                 dbtype=dbtype)
        else:
            raise Exception('dbtype is incorrect')


    def tearDown(self):
        self.monsql.close()


#encoding=utf-8

"""
This is a module for using mysql as simple as using mongodb
Support mongodb manipulation like:

* find(queryObj, fieldObject)
* find_one(queryObj, fieldObject)
* insert(keyValueObj)
* insert_batch(a list of keyValueObj)
* update(queryObj, keyValueObj)
* remove(queryObj)

The design for this module would take **Pymongo** as a reference

Since it's a module for mysql, transaction must be considered, so we also provide:
* commit(): commit the transaction

We plan to Support Table join
We would also support subquery
"""

from config import TRANSACTION_MODE
from exception import MonSQLException
from table import Table
from queryset import DataRow
import abc

class Database:
    """
    MongoDB style of using Relational Database

    :Examples:

    >>> monsql = MonSQL(host, port, username, password, dbname, DBTYPES.MYSQL)
    >>> user_table = monsql.get('user')
    >>> activated_users = user_table.find({'state': 2})
    >>> user_ids = user_table.insert([{'username': ...}, {'username': ...}, ...])
    >>> user_table.commit() # OR monsql.commit()

    """

    def __init__(self, db, mode=TRANSACTION_MODE.DEFAULT):
        self.__db = db
        self.__cursor = self.__db.cursor()
        self.__table_map = {}
        self.__mode = mode

    """
    Properties for accessibility to subclasses
    """
    @property
    def cursor(self):
        return self.__cursor

    @property
    def db(self):
        return self.__db

    @property
    def mode(self):
        return self.__mode

    def __ensure_table_obj(self, name):
        if not self.__table_map.has_key(name):
            self.__table_map[name] = self.get_table_obj(name)

    @abc.abstractmethod
    def get_table_obj(self, name):
        """This is used internally inside the class
        Implemented by subclasses, because different database may use different table class"""
        pass

    @abc.abstractmethod
    def list_tables(self):
        """
        Return a list of lower case table names. Different databases have their own ways to 
        do this, so leave the implementation to the subclasses
        """
        pass

    @abc.abstractmethod
    def truncate_table(self, tablename):
        """Delete all rows in a table. 
        Not all databases support built-in truncate, so implementation is left
        to subclasses. For those don't support truncate, 'delete from ...' is used """
        pass

    def get(self, name):
        """
        Return a Table object to perform operations on this table. 

        Note that all tables returned by the samle Database instance shared the same connection.

        :Parameters:

        - name: A table name

        :Returns: A Table object
        """
        self.__ensure_table_obj(name)
        return self.__table_map[name]

    def close(self):
        """
        Close the connection to the server
        """
        self.__db.close()
        self.__table_map = {}

    def commit(self):
        """
        Commit the current session
        """
        self.__db.commit()
    
    def set_foreign_key_check(self, to_check):
        """
        Enable/disable foreign key check. Disabling this is especially useful when
        deleting from a table with foreign key pointing to itself
        """
        if to_check:
            self.__db.cursor().execute('SET foreign_key_checks = 1;')
        else:
            self.__db.cursor().execute('SET foreign_key_checks = 0;')

    def is_table_existed(self, tablename):
        """
        Check whether the given table name exists in this database. Return boolean.
        """
        all_tablenames = self.list_tables()
        tablename = tablename.lower()

        if tablename in all_tablenames:
            return True
        else:
            return False

    def create_table(self, tablename, columns, primary_key=None, force_recreate=False):
        """
        :Parameters:

        - tablename: string
        - columns: list or tuples, with each element be a string like 'id INT NOT NULL UNIQUE'
        - primary_key: list or tuples, with elements be the column names
        - force_recreate: When table of the same name already exists, if this is True, drop that table; if False, raise exception
        
        :Return: Nothing

        """
        if self.is_table_existed(tablename):
            if force_recreate:
                self.drop_table(tablename)
            else:
                raise MonSQLException('TABLE ALREADY EXISTS')

        columns_specs = ','.join(columns)
        if primary_key is not None:
            if len(primary_key) == 0:
                raise MonSQLException('PRIMARY KEY MUST AT LEAST CONTAINS ONE COLUMN')

            columns_specs += ',PRIMARY KEY(%s)' %(','.join(primary_key))

        sql = 'CREATE TABLE %s(%s)' %(tablename, columns_specs)
        self.__cursor.execute(sql)
        self.__db.commit()

    def drop_table(self, tablename, silent=False):
        """
        Drop a table

        :Parameters:

        - tablename: string
        - slient: boolean. If false and the table doesn't exists an exception will be raised;
          Otherwise it will be ignored
        
        :Return: Nothing

        """
        if not silent and not self.is_table_existed(tablename):
            raise MonSQLException('TABLE %s DOES NOT EXIST' %tablename)

        self.__cursor.execute('DROP TABLE IF EXISTS %s' %(tablename))
        self.__db.commit()


    def raw(self, sql):
        """
        Execute raw sql
        :Parameters:

        - sql: string, sql to be executed

        :Return: the result of this execution

        If it's a select, return a list with each element be a DataRow instance

        Otherwise return raw result from the cursor (Should be insert or update or delete)

        """
        res = self.cursor.execute(sql)
        if self.cursor.description is None:
            return res

        rows = self.cursor.fetchall()
        columns = [d[0] for d in self.cursor.description]

        structured_rows = []
        for row in rows:
            data = {}
            for val, col in zip(row, columns):
                data[col] = val
            structured_rows.append(DataRow(data))

        return structured_rows

    




from setuptools import setup, find_packages

setup(name='monsql',
      version='0.1.7',
      packages = find_packages(),
      author='firstprayer',
      author_email='zhangty10@gmail.com',
      description='MonSQL - Mongodb-style way for using mysql.',
      url='https://github.com/firstprayer/monsql.git',
      install_requires=[
      	'MySQL-python'
      ],
)
